# Force Shield Modification, (c) 2007 Zodius Infuser
# Licensed under GPL v2 or later

local oldShield = Shield

Shield = Class(oldShield){
	OnCreate = function(self,spec)
		self.Owner = spec.Owner

		--Set shield constants
		self:SetThreshold(0.01) --Amount by which a shield needs to change in size before it updates
		self:SetEnergyFactor(1.3) -- The amount of energy used by a weapon in comparison to the default energy
		self:SetTimeDelay(0.5) -- The time delay on threads
		
		--Set all Default Values
		self:SetDefaultSize(spec.Size)
		self:SetDefaultVerticalOffset(spec.ShieldVerticalOffset)
		self:SetDefaultEnergyUsage(self.Owner:GetBlueprint().Economy.MaintenanceConsumptionPerSecondEnergy)
		
		--Set all Minimum Values
		self:SetMinSize(self:CalcMinSize())
		self:SetMinVerticalOffset(0)
		
		--Set all Current Values
		self:SetSize(self:CalcSize())
		self:SetVerticalOffset(self:CalcVerticalOffset())
		self:SetEnergyUsage(self.DefaultEnergyUsage)
		
		oldShield.OnCreate(self,spec)
	end,



	##########################################################################################
	## SET FUNCTIONS
	##########################################################################################		

	SetDefaultSize = function(self, size)
		self.DefaultSize = size
	end,
	
	SetMinSize = function(self, size)
		self.MinSize = size
	end,
	
	SetDefaultVerticalOffset = function(self, offset)
        self.ShieldDefaultVerticalOffset = offset
    end,
	
	SetMinVerticalOffset = function(self, offset)
        self.ShieldMinVerticalOffset = offset
    end,
	
	SetDefaultEnergyUsage = function(self, energy)
		self.DefaultEnergyUsage = energy
	end,
	
	SetEnergyUsage = function(self, energy)
		self.EnergyUsage = energy
	end,
	
	SetThreshold = function(self, threshold)
		self.Threshold = threshold
	end,
	
	SetEnergyFactor = function(self, factor)
		self.EnergyFactor = factor
	end,

	SetTimeDelay = function(self, delay)
		self.TimeDelay = delay
	end,

	
	##########################################################################################
    ## CALCULATION FUNCTIONS
    ##########################################################################################

	#
	# Calculates the minimum size that the shield needs to be in order to protect the shield owner
	# Uses sizes from the owners blueprint to determine this size
	#
	CalcMinSize = function(self)
		local x = self.Owner:GetBlueprint().Physics.MeshExtentsX or 6
		local y = self.Owner:GetBlueprint().Physics.MeshExtentsY or 6
		local z = self.Owner:GetBlueprint().Physics.MeshExtentsZ or 6
		#local x, y, z = self.Owner:GetUnitSizes()
	    return math.sqrt(math.pow(x, 2) + math.pow(y, 2) + math.pow(z, 2))
	end,
	
	#
	# Calculates the current size that the shield should be, based on the current shield health
	#
	CalcSize = function(self)
		local value = math.pow(self:GetHealth() / self:GetMaxHealth(), 1 / 3) * (self.DefaultSize - self.MinSize)
		return value + self.MinSize
	end,
	
	#
	# Calculates the current vertical offset that the shield should have, based on the current shield health
	#
	CalcVerticalOffset = function(self)
		local value = math.pow(self:GetHealth() / self:GetMaxHealth(), 1 / 3) * (self.ShieldDefaultVerticalOffset - self.ShieldMinVerticalOffset)
		return value + self.ShieldMinVerticalOffset
	end,
	
	#
	# The weapon energy is based on the amount of damage caused by it in comparison to the maximum shield health
	# This is calculated as a percentage of the default energy usage, which is the multiplied by an energy factor,
	# making the weapon use more or less energy than the calculated percentage
	#
	CalcWeaponEnergy = function(self, amount)
		#return (self.DefaultEnergyUsage * (amount / self:GetMaxHealth())) * self.EnergyFactor
		return math.pow(self.DefaultEnergyUsage, self.EnergyFactor) * (amount / self:GetMaxHealth())
	end,
	
	
	
	##########################################################################################
    ## SHIELD FUNCTIONS
    ##########################################################################################
	
	#
	# Updates the shield render without deactivating and reactivating it
	#
	UpdateShieldMesh = function(self)
		self:SetCollisionShape('Sphere', 0, 0, 0, self.Size / 2)
		self:SetParentOffset(Vector(0,self.ShieldVerticalOffset,0))
		self:SetDrawScale(self.Size)
		self.MeshZ:SetDrawScale(self.Size)
		self.MeshZ:SetParentOffset(Vector(0,self.ShieldVerticalOffset,0))
	end,

	#
	# Creates the shield at the size and vertical offset proportional to the shields health
	#
    CreateShieldMesh = function(self)
		self:SetVerticalOffset(self:CalcVerticalOffset())
		self:SetSize(self:CalcSize())
		oldShield.CreateShieldMesh(self)
    end,

	#
	# Modified version of OnCollisionCheck which removes the checking for nukes.
	# Nukes can now impact the shield rather than being unaffected by it.
	#
    OnCollisionCheck = function(self,other)
        if other:GetArmy() == -1 then
            return false
        end
		
        if other:GetBlueprint().Physics.CollideFriendlyShield then
            return true
        end
		
        return IsEnemy(self:GetArmy(),other:GetArmy())
    end,
	
	#
	# Damage caused by weapon varies according to the energy and size of a shield
	# All values are in comparison with their defaults, allowing more efficient shields to be produced
	# The energy / the volume gives the shield density which the weapon amount is divided by
	#
	# Original Equation: (Energy used by Shield / Size of Shield)/ Surface Area of Shield --Size ^ 2
	# New Equation: Energy used by Shield / Volume of Shield --Size ^ 3
	#
    OnGetDamageAbsorption = function(self,instigator,amount,type)
		local energy = self.EnergyUsage / self.DefaultEnergyUsage
		local distance = self.Size / self.DefaultSize
		local particles = energy / distance
		local area =  math.pow(self.Size / self.DefaultSize, 2)
		local density = particles / area
        return math.min(self:GetHealth(), (amount / density))
    end,
	
    OnDamage = function(self,instigator,amount,vector,type)
		self:SetEnergyUsage(self:CalcWeaponEnergy(amount) + self.EnergyUsage)
		local absorbed = self:OnGetDamageAbsorption(instigator,amount,type)
        self:AdjustHealth(instigator, -absorbed)

		if not self.ModifyThread then
			self.ModifyThread = ForkThread(self.ModifyShieldThread, self)
			self.Owner.Trash:Add(self.ModifyThread)
		end
		
		if not self.EnergyThread then
			self.EnergyThread = ForkThread(self.EnergyUsageThread, self)
			self.Owner.Trash:Add(self.EnergyThread)
		end

		--Copied from OnDamage in shield.lua, Can't override due to actions needing to be performed after self:AdjustHealth()
		if not self.Owner:IsUnitState('Building') and
           not self.Owner:IsUnitState('Enhancing') and
           not self.Owner:IsUnitState('Repairing') then
            self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
        end

        #LOG('Shield Health: ' .. self:GetHealth())
        if self.RegenThread then
           KillThread(self.RegenThread)
           self.RegenThread = nil
        end
        if self:GetHealth() <= 0 then
            ChangeState(self, self.DamageRechargeState)
        else
            ForkThread(self.CreateImpactEffect, self, vector)
            if self.RegenRate > 0 then
                self.RegenThread = ForkThread(self.RegenStartThread, self)
                self.Owner.Trash:Add(self.RegenThread)
            end
        end
    end,
	
    RegenStartThread = function(self)
        WaitSeconds(self.RegenStartTime)
        while self:GetHealth() < self:GetMaxHealth() do
            self:AdjustHealth(self.Owner, self.RegenRate)
			
			if not self.ModifyThread then
				self.ModifyThread = ForkThread(self.ModifyShieldThread, self)
				self.Owner.Trash:Add(self.ModifyThread)
			end
			
			--Copied from RegenStartThread in shield.lua, Can't override due to actions needing to be performed after self:AdjustHealth()
            if not self.Owner:IsUnitState('Building') and
               not self.Owner:IsUnitState('Enhancing') and
               not self.Owner:IsUnitState('Repairing') then
                self.Owner:SetWorkProgress(self:GetHealth() / self:GetMaxHealth())
            end

            WaitSeconds(1)
        end
    end,
	
	
	
	##########################################################################################
    ## FORCE SHIELD THREADS
    ##########################################################################################

	#
	# Regularly modifies the size of the shield until it returns to the default size
	#
	ModifyShieldThread = function(self)
		while (self:CalcSize() <= (self.Size - self.Threshold)) or (self:CalcSize() >= (self.Size + self.Threshold)) do
			--Set size and vertical offset to the mean of their old and new values
			self:SetVerticalOffset(self:CalcVerticalOffset())
			self:SetSize(self:CalcSize())
			
			--Update the shield mesh
			if self:IsOn() then
				self:UpdateShieldMesh()
			end
			WaitSeconds(self.TimeDelay)
		end
		self.ModifyThread = nil
	end,
	
	#
	# Regularly updates the energy drain of the owner to match the changing weapon energy
	#
	EnergyUsageThread = function(self)
		while self.EnergyUsage > self.DefaultEnergyUsage do
			--Set the energy maintenance consumption override of the owner
			self.Owner:SetEnergyMaintenanceConsumptionOverride(self.EnergyUsage)
			self.Owner:UpdateConsumptionValues()
			#LOG('Energy Usage: ' .. self.EnergyUsage)
			self:SetEnergyUsage(self.DefaultEnergyUsage)
			
			WaitSeconds(self.TimeDelay)
		end
		self.Owner:SetEnergyMaintenanceConsumptionOverride(self.DefaultEnergyUsage)
		self.Owner:UpdateConsumptionValues()
		self:SetEnergyUsage(self.DefaultEnergyUsage)
		self.EnergyThread = nil
	end,
}
